// ReadHVs.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
/*
#include <sys/timeb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <process.h>
*/
#ifdef WIN32
#include <windows.h>
#include <sys/timeb.h>
#endif
#include <io.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

static unsigned char *bobsub;

#define MAXNODES 8192
static char *seps = " ,;\t\n";
static int nsub, node0;
static __int64 tile_off64t[MAXNODES];
static int tile_nrep[MAXNODES];
static int tile_nxsub[MAXNODES], tile_nysub[MAXNODES], tile_nzsub[MAXNODES];
static int tile_ixoff[MAXNODES], tile_iyoff[MAXNODES], tile_izoff[MAXNODES];
static int tile_parent[MAXNODES];
static int tile_maxerr[MAXNODES];
static int ncPath;
static char tile_file[MAXNODES][64], tile_path[64], tile_full[256];
static char line[256];
static int tile_children[MAXNODES][32];
static float fLimits[6];

int main(int argc, char* argv[])
{
    int match( char *token, char *word, int tokenlength );
	DWORD dhpReadFromFileSync(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);
	DWORD nbytes, ngot;
	char cFile[256];
	int iNBoundariesRep[129];
	int nxsub, nysub, nzsub, nxsb2, nysb2, nzsb2;
	int i, ixoff, iyoff, izoff, isub;
	SIZE_T dwNeedSize, dwMaxSize;
	FILE *fpp;
	int fd;

	if(argc != 5) {
		printf("ReadHVs reads 1 block from an HV file and writes it to a BOB file\n");
		printf("Usage: ReadHVs HVfileName ixoff iyoff izoff\n");
		return 0;
	}


	// Parse comand line arguments
	sscanf(argv[1], "%s", cFile);    printf("Input HV file: %s\n", cFile);
	sscanf(argv[2], "%d", &ixoff);   printf("X-offset     : %d\n", ixoff);
	sscanf(argv[3], "%d", &iyoff);   printf("Y-offset     : %d\n", iyoff);
	sscanf(argv[4], "%d", &izoff);   printf("Z-offset     : %d\n", izoff);


	// Read and parse header info
	fpp = fopen(cFile, "r");
	if(fpp == NULL) {
		printf("Could not open %s\n", cFile);
		return 0;
	}

	dwMaxSize = (SIZE_T)(128*128*128);

	for(i=0; i<129; i++) { iNBoundariesRep[i] = 0; }
	nsub = 0;
	node0 = -1;
	while (node0 < 0) {
		i = match(fgets(line,200,fpp), "end", 3);
		strncpy(tile_file[nsub], strtok( line,seps ), 128);
		sscanf(strtok(NULL, seps), "%I64d", &tile_off64t[nsub]);
		tile_nrep[nsub]   = atoi( strtok(NULL,seps) );
		tile_nxsub[nsub]  = atoi( strtok(NULL,seps) );
		tile_nysub[nsub]  = atoi( strtok(NULL,seps) );
		tile_nzsub[nsub]  = atoi( strtok(NULL,seps) );
		tile_ixoff[nsub]  = atoi( strtok(NULL,seps) );
		tile_iyoff[nsub]  = atoi( strtok(NULL,seps) );
		tile_izoff[nsub]  = atoi( strtok(NULL,seps) );
		tile_parent[nsub] = atoi( strtok(NULL,seps) );
		tile_maxerr[nsub] = atoi( strtok(NULL,seps) );
		if(tile_parent[nsub] < 0) { node0 = nsub; }

		if( iNBoundariesRep[tile_nrep[nsub]] < -tile_ixoff[nsub] ) {
			iNBoundariesRep[tile_nrep[nsub]] = -tile_ixoff[nsub];
		}

		// Get Max memory needy for any brick in this hv file
		// This wiil get used below to determine if bob arrays need to be resized
		nxsub = tile_nxsub[nsub];
		nysub = tile_nysub[nsub];
		nzsub = tile_nzsub[nsub];
		nxsb2 = 2;   while(nxsb2 < nxsub) { nxsb2 *= 2; }
		nysb2 = 2;   while(nysb2 < nysub) { nysb2 *= 2; }
		nzsb2 = 2;   while(nzsb2 < nzsub) { nzsb2 *= 2; }
		dwNeedSize = (SIZE_T)(nxsb2 * nysb2 * nzsb2);
		if(dwMaxSize < dwNeedSize) { dwMaxSize = dwNeedSize; }

		nsub++;
	}
	fclose(fpp);


	// DEBUG
	if(0) {
		printf("number of nodes: nsub = %d\n", nsub);
		printf("max size of a block dwMaxSize = %d\n", (int)dwMaxSize);
		printf("\n\nNODE NREP      XYZsize     XYZoffset\n");
		for(i=0; i<nsub; i++) {
			printf(" %3d  %3d %4d%4d%4d  %4d%4d%4d\n", i, tile_nrep[i],
				tile_nxsub[i], tile_nysub[i], tile_nzsub[i],
				tile_ixoff[i], tile_iyoff[i], tile_izoff[i]   );
		}
	}


	// Find fill res. node (nrep=1) which contains (ixoff,iyoff,izoff)
	isub = -1;
	for(i=0; i<nsub; i++) {
		if(    tile_nrep[i] == 1
			&& tile_ixoff[i] < ixoff  &&  ixoff < tile_ixoff[i]+tile_nxsub[i]
			&& tile_iyoff[i] < iyoff  &&  iyoff < tile_iyoff[i]+tile_nysub[i]
			&& tile_izoff[i] < izoff  &&  izoff < tile_izoff[i]+tile_nzsub[i] ) {

			isub = i;
		}
	}
	if(isub < 0) {
		printf("Did not find any node which contains voxel %d %d %d\n",
			ixoff, iyoff, izoff);
		return 0;
	}
	printf("\nFound isub = %d\n", isub);


	// Read in block
	bobsub = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
	nbytes = (DWORD)(tile_nxsub[isub]*tile_nysub[isub]*tile_nzsub[isub]);
	ngot = dhpReadFromFileSync(bobsub, cFile, tile_off64t[isub], nbytes);


	// Write block to a file
	fd = _open("one.bob", _O_WRONLY | _O_CREAT | _O_BINARY , _S_IREAD | _S_IWRITE );
	write(fd, bobsub, (int)nbytes);
	_close(fd);


	return 0;
}

int match( char *token, char *word, int tokenlength ) {
  int n,i;
  n = strlen(word);
  if ( n>tokenlength ) n= tokenlength;
  i = !strncmp(token,word,n);
  return i;
}

DWORD dhpReadFromFileSync(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes)
{
	HANDLE hFile;                        // handle to read file
	__int64 shift32, shift8;
	LONG dwHigh, dwLow;
	unsigned long ngot;

	shift8  = (__int64)256;
	shift32 = shift8 * shift8 * shift8 * shift8;
	dwHigh  = (LONG)(offset / shift32);
	dwLow   = (LONG)(offset % shift32);
	hFile = CreateFile(lpFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL, NULL);
	SetFilePointer(hFile, dwLow, &dwHigh, FILE_BEGIN);
	ReadFile(hFile, lpBuffer, nbytes, &ngot, NULL);
	CloseHandle(hFile);

	return ngot;
}

